from typing import Any, Dict, List, Optional, Union

from pandas import DataFrame

from .._base import _logger
from . import _decorator
from ._config import _Config
from ._data_objects import DataPoints, Investments
from ._investment._peer_group_data import PeerGroupProvider
from .data_type import Order, PeerGroupMethodology

_config = _Config()
peer_group_provider = PeerGroupProvider()


@_decorator.typechecked
def get_peer_group_breakpoints(
    investments: Union[List[str], str],
    data_points: Union[List[Dict[str, Any]], DataFrame],
    order: Order = Order.ASC,
    percentiles: Optional[List[int]] = list(range(1, 101)),
    methodology: Optional[PeerGroupMethodology] = None,
) -> DataFrame:
    """Returns peer group breakpoints for the specified list of investments and data points.

    Args:
        investments (:obj:`Union`, `required`): Defines the investments to fetch. Input can be:

            * Investment IDs (:obj:`list`, `optional`): Investment identifiers, in the format of SecId;Universe or just SecId. E.g., ["F00000YOOK;FO","FOUSA00CFV;FO"] or ["F00000YOOK","FOUSA00CFV"]. Use the `investments <./lookup.html#morningstar_data.direct.investments>`_ function to discover identifiers.
            * Investment List ID (:obj:`str`, `optional`): Saved investment list in Morningstar Direct. Use the `get_investment_lists <./lists.html#morningstar_data.direct.user_items.get_investment_lists>`_ function to discover saved lists.
            * Search Criteria  ID (:obj:`str`, `optional`): Saved search criteria in Morningstar Direct. Use the `get_search_criteria <./search_criteria.html#morningstar_data.direct.user_items.get_search_criteria>`_ function to discover saved search criteria.
            * Search Criteria Condition (:obj:`dict`, `optional`): Search criteria definition. See details in the Reference section of `get_investment_data <./investment.html#morningstar_data.direct.get_investment_data>`_ or use the `get_search_criteria_conditions <./search_criteria.html#morningstar_data.direct.user_items.get_search_criteria_conditions>`_ function to discover the definition of a saved search criteria.

        data_points (:obj:`Union`, `optional`): Defines the data points to fetch. If not provided and investments are specified with a list ID or search criteria ID, the corresponding bound dataset will be used.

            * Data Point IDs (:obj:`List[Dict]`, `optional`): A list of dictionaries, each defining a data point and its (optional) associated settings. The optional `alias` attribute can be added to each data point and will be used in the response, e.g., [{"datapointId": "41", "alias": "Z1"}. Use `get_data_set_details <./data_set.html#morningstar_data.direct.user_items.get_data_set_details>`_ to discover data point identifiers from a saved data set.
            * Data Point Settings (:obj:`DataFrame`, `optional`): A DataFrame of data point identifiers and their associated settings. Use the `get_data_set_details <./data_set.html#morningstar_data.direct.user_items.get_data_set_details>`_ function to discover data point settings from a saved data set.

        order (:obj:`md.direct.data_type.Order`, `optional`): Breakpoint order, can be set to ``md.direct.data_type.Order.DESC`` (descending) or ``md.direct.dats_type.Order.ASC`` (ascending, default).
        percentiles (:obj:`list`, `optional`): Percentiles default to a list [1,2,3,...,100] if not provided, values should be within 1-100 range.
        methodology (:obj:`md.direct.data_type.PeerGroupMethodology`, `optional`): Breakpoint calculation methodology, can be set to ``md.direct.data_type.PeerGroupMethodology.MORNINGSTAR`` or ``md.direct.data_type.PeerGroupMethodology.SIMPLE``. Defaults to the global setting "Custom Peer Group Ranking" in Morningstar Direct if not provided.

    :Returns:
        DataFrame: A DataFrame object with peer group breakpoint data. The columns include the `alias` that the user
        input in the `data_points` parameter.

    :Examples:

    Get peer group breakpoint data for the standard deviation data point.

    ::

        import morningstar_data as md

        df = md.direct.get_peer_group_breakpoints(
                investments='740284aa-fcd3-43f6-99d1-8f3d4a179fcc',
                data_points=[
                    {"datapointId": "41", "alias": "Z1"},
                    {"datapointId": "41", "alias": "Z2", "startDate": "2021-07-01", "endDate": "2021-12-31", "windowType": "2", "windowSize": "3", "stepSize": "2"}
                ],
                order=md.direct.data_type.Order.ASC,
                percentiles=[25, 50, 75, 100]
            )
        df

    :Output:
        ======  ===========  ==========  ==========  ==========  ==========  ==========
        Alias   StartDate    EndDate     25          50          75          100
        ======  ===========  ==========  ==========  ==========  ==========  ==========
        Z1      2019-04-01   2022-03-31  17.301437   12.720889   7.055372    -3.460187
        Z2      2021-07-01   2021-09-30  1.827371    -0.804269   -4.899745   -52.143678
        Z2      2021-09-01   2021-11-30  -0.030321   -4.336051   -10.618009  -40.980480
        ======  ===========  ==========  ==========  ==========  ==========  ==========

    :Errors:
        AccessDeniedError: Raised when the user lacks permission or is not authorized to access the resource.

        BadRequestException: Raised due to invalid/incorrect request, malformed request syntax, or deceptive request routing.

        NetworkExceptionError: Raised when there is an issue with the internet connection or if the request is made from an unsecure network.

        ResourceNotFoundError: Raised when the requested resource does not exist in Direct.

    """
    investment_param = Investments(investments)
    data_point_param = DataPoints(data_points)

    peer_group_req = peer_group_provider.build_request(
        investment_param, data_point_param, order=order, percentiles=percentiles, methodology=methodology
    )
    peer_group_resp = peer_group_provider.run_request(peer_group_req)
    return peer_group_provider.build_data_frame(peer_group_resp)



from dataclasses import dataclass
from typing import Optional


class MdBaseException(Exception):
    pass


@dataclass
class MdApiRequestException(MdBaseException):
    request_id: str
    url: str
    detail: str

    def __str__(self) -> str:
        output = f"MdApiRequestException\nURL: {self.url}\nRequest ID: {self.request_id}"
        if self.detail:
            output += f"\nDetail: {self.detail}"
        return output


@dataclass
class MdApiTaskException(MdBaseException):
    task_id: str
    detail: Optional[str]

    def __str__(self) -> str:
        output = f"\nTask ID: {self.task_id}"
        if self.detail:
            output += f"\nDetail: {self.detail}"
        return output


@dataclass
class MdApiTaskTimeoutException(MdBaseException):
    task_id: str
    timeout_time: int

    def __str__(self) -> str:
        return f"Task timed out with timeout of {self.timeout_time} seconds.\nTask ID: {self.task_id}"


@dataclass
class MdApiTaskFailure(MdBaseException):
    message: str
    request_id: Optional[str] = None
    task_id: Optional[str] = None

    def __str__(self) -> str:
        if self.task_id:
            output = f"\nTask ID: {self.task_id}"
        else:
            output = f"\nRequest ID: {self.request_id}"
        if self.message:
            output += f"\nDetail: {self.message}"
        return output


class BadRequestError(MdApiTaskFailure):
    pass


class ResourceNotFoundError(MdApiTaskFailure):
    pass


class AccessDeniedError(MdApiTaskFailure):
    pass


# We may get exceptions passed to us from the API in string (json) format. At this point we will have
# the exception class name, the error message, and either a task id for request id.
# This function is used to create an exception object based on these values.
def exception_by_name(
    name: str, message: str = "", task_id: Optional[str] = None, request_id: Optional[str] = None
) -> MdApiTaskFailure:
    # Let's see if the exception is a subclass of MdApiTaskFailure.
    # If it is, we'll use that class to create the exception object. Otherwise default to the base MdApiTaskFailure class.
    throwable_exceptions = MdApiTaskFailure.__subclasses__()
    exc_class = next((exc for exc in throwable_exceptions if exc.__name__ == name), MdApiTaskFailure)

    return exc_class(task_id=task_id, request_id=request_id, message=message)



investments_list_default_columns = [
    "secid",
    "masterportfolioid",
    "tradingsymbol",
    "name",
    "securitytype",
    "exchangeid",
    "category",
]

# these datapointIds correspond to the above columns
investments_list_data_points = [
    {"datapointId": "OS00I", "isTsdp": False},
    {"datapointId": "DC09A", "isTsdp": False},
    {"datapointId": "OS385", "isTsdp": False},
    {"datapointId": "OS01W", "isTsdp": False},
    {"datapointId": "OS010", "isTsdp": False},
    {"datapointId": "OS01Z", "isTsdp": False},
    {"datapointId": "OS245", "isTsdp": False},
]



from .._base import _logger
from ..mdapi import RequestObject
from . import user_items
from ._config import _Config
from ._portfolio_data_set import PortfolioDataSet
from .asset_flow import (
    get_asset_flow,
    get_asset_flow_data_points,
    get_asset_flow_markets,
)
from .custom_database import firm_database, my_database
from .data_type import Blank, Frequency, Order, PeerGroupMethodology, TimeSeriesFormat
from .holdings import (
    get_holding_dates,
    get_holdings,
    get_lookthrough_holdings,
    holding_dates,
    holdings,
)
from .investment import get_investment_data, investment_data
from .lookup import (
    companies,
    firms,
    get_brandings,
    get_data_point_settings,
    get_morningstar_data_sets,
    investment_universes,
    investments,
    portfolio_managers,
)
from .peer_group import get_peer_group_breakpoints
from .performance_report import (
    calculate_report,
    get_report,
    get_report_status,
    get_reports,
)
from .returns import excess_returns, get_excess_returns, get_returns, returns

from . import portfolio  # isort: skip -> to avoid circular import error this needs to be at the end.


config = _Config()



"""
Public method parameter and return types
"""
from __future__ import annotations

from enum import Enum

from ._data_type import RaiseEnum


class Universe:
    def __init__(self, n: str, u: str):
        self.name = n
        self.universe_id = u


class Frequency(str, Enum, metaclass=RaiseEnum):
    def __new__(cls, value: str, data_point_id: str, abbr: str) -> Frequency:
        obj = str.__new__(cls, value)
        obj._value_ = value
        return obj

    def __init__(self, value: str, data_point_id: str, abbr: str) -> None:
        self.data_point_id = data_point_id
        self.abbr = abbr

    daily = ("daily", "HS793", "d")
    weekly = ("weekly", "HP002", "w")
    monthly = ("monthly", "HP010", "m")
    quarterly = ("quarterly", "HP020", "q")
    yearly = ("yearly", "HS803", "a")


class Blank(str, Enum, metaclass=RaiseEnum):
    warning = ("warning",)
    ignore = ("ignore",)
    update = ("update",)

    @classmethod
    def options(cls) -> str:
        return ", ".join(cls._member_map_.keys())


class PeerGroupMethodology(str, Enum):
    MORNINGSTAR = ("MORNINGSTAR",)
    SIMPLE = ("SIMPLE",)


class Order(str, Enum):
    ASC = ("ASC",)
    DESC = ("DESC",)


class TimeSeriesFormat(str, Enum):
    LONG = "Long"
    WIDE = "Wide"
    LONG_WITHOUT_NAME = "LongWithoutName"


# Calculation window type
class CalculationWindowType(str, Enum):
    SINGLE_DATA_POINT = ("1",)
    ROLLING_WINDOW = ("2",)
    FORWARD_EXTENDING_WINDOW = ("3",)
    BACKWARD_EXTENDING_WINDOW = ("4",)



import logging
import os
from typing import Any, Dict, List, Union

import polling2
import simplejson as json
from boto3.s3.transfer import TransferConfig
from botocore.exceptions import ClientError

from ..direct import _decorator
from ..direct._backend_apis import DeliveryAPIBackend
from ..direct._error_messages import FAILED_DELIVERY_ERROR
from ..direct._exceptions import FailedDeliveryError
from ._delivery_config import DeliveryConfig, DeliveryType
from ._helpers import (
    _create_feed_body,
    _create_output_file_name,
    _encrypt_ftp_password,
    _get_bucket_name,
    _get_job_name,
    _get_mds_base_url,
    _get_mds_bucket,
    _get_user_id,
)

_logger = logging.getLogger(__name__)
_api_backend = DeliveryAPIBackend()


def _create_feed_id() -> int:
    """Calls the MDS feeds api to generate a new feed id.

    This feed id is needed when calling the MDS deliver api. In general, when a notebook is running
    in a schedule it will be passed in this feed id. But for an ad-hoc delivery this feed id needs
    to be created before the MDS delivery api can be called.
    """
    _logger.info("Creating a feed id using MDS api")
    url = f"{_get_mds_base_url()}v1/feeds?type=notebook"
    data = _create_feed_body()

    try:
        _logger.info(f"Calling MDS /feeds endpoint with: {data}")
        response_json = _api_backend.do_post_request(url, data=json.dumps(data, ignore_nan=True))
        feed_id: int = response_json["feedId"]
        return feed_id
    except Exception:
        _logger.error("Could not create a feed id using MDS api")
        raise


def _upload_to_s3(file_path: str, output_file_name: str) -> None:
    """Uploads a file to the MDS s3 bucket"""
    try:
        config = TransferConfig(
            multipart_threshold=5 * 1024 * 1024 * 1024,  # 5GB
            max_concurrency=10,
            multipart_chunksize=500 * 1024 * 1024,  # 500MB
            use_threads=True,
        )

        output_bucket = _get_mds_bucket()
        output_bucket.upload_file(file_path, output_file_name, Config=config)
    except FileNotFoundError:
        _logger.error(f"File not found: {file_path}, make sure the file you are trying to upload exists")
        raise
    except ClientError:
        _logger.error("Could not upload to s3 bucket due to AWS error")
        raise
    except Exception:
        _logger.error("Could not upload to s3 bucket")
        raise


def _process_configs(configs: List[DeliveryConfig]) -> Dict[str, List]:
    unique_files = set()
    unique_emails = set()
    unique_ftps = set()
    unique_profiles = set()

    file_mappings = []

    # Processing each config
    for config in configs:
        if config.method == DeliveryType.EMAIL:
            unique_emails.add(config.email_address)
        elif config.method == DeliveryType.DELIVERY_PROFILE:
            unique_profiles.add(config.delivery_profile_id)
        else:
            unique_ftps.add((config.user_name, config.password, config.server, config.folder))
        unique_files.add((config.file_path, config.delivered_file_name, config.include_timestamp))

    # Converting sets to lists to maintain order and for indexing
    unique_files_list = list(unique_files)
    unique_emails_list = list(unique_emails)
    unique_ftps_list = list(unique_ftps)
    unique_profiles_list = list(unique_profiles)

    # Creating mappings
    for config in configs:
        file_index = unique_files_list.index((config.file_path, config.delivered_file_name, config.include_timestamp))
        if config.method == DeliveryType.EMAIL:
            email_index = unique_emails_list.index(config.email_address)
            mapping = {"fileIndex": f"{file_index}", "deliveryInfo": f"emails.{email_index}"}
        elif config.method == DeliveryType.DELIVERY_PROFILE:
            profile_index = unique_profiles_list.index(config.delivery_profile_id)
            mapping = {"fileIndex": f"{file_index}", "deliveryInfo": f"profiles.{profile_index}"}
        else:
            ftp_index = unique_ftps_list.index((config.user_name, config.password, config.server, config.folder))
            mapping = {"fileIndex": f"{file_index}", "deliveryInfo": f"ftps.{ftp_index}"}

        file_mappings.append(mapping)

    # Remove duplicate mappings
    file_mappings = [dict(t) for t in {tuple(d.items()) for d in file_mappings}]

    files_list = [
        {
            "key": file[0],
            "fileName": file[1],
            "bucket": _get_bucket_name(),
            "includeTimestamp": str(file[2]),
        }
        for file in unique_files_list
    ]
    emails_list = [{"address": email} for email in unique_emails_list]
    profiles_list = [{"deliveryProfileId": profile} for profile in unique_profiles_list]
    ftps_list = [
        {"username": ftp[0], "password": _encrypt_ftp_password(ftp[1]), "server": ftp[2], "folder": ftp[3]}
        for ftp in unique_ftps_list
    ]

    return {
        "files": files_list,
        "emails": emails_list,
        "profiles": profiles_list,
        "ftps": ftps_list,
        "delivery_details": file_mappings,
    }


def _create_deliver_body(configs: List[DeliveryConfig]) -> Dict:
    """Creates the post body for calling  MDS delivery API with single or multiple files/delivery destinations."""

    # convert list of configs to mds post body format
    config_mapping = _process_configs(configs=configs)

    body = {
        "userId": _get_user_id(),
        "userName": _get_user_id(),
        "jobName": _get_job_name(),
        "files": config_mapping["files"],
        "deliveryInfos": {
            "emails": config_mapping["emails"],
            "profiles": config_mapping["profiles"],
            "ftps": config_mapping["ftps"],
        },
        "deliveryDetails": config_mapping["delivery_details"],
    }
    return body


def _mds_deliver_file(config: List[DeliveryConfig], feed_id: int) -> str:
    """Calls the MDS delivery api with the needed information to deliver the uploaded file and returns job_id"""
    url = f"{_get_mds_base_url()}v1/delivery/{feed_id}"
    data = _create_deliver_body(config)
    _logger.info(f"Calling MDS /delivery/{feed_id} endpoint with: {data}")
    response_json = _api_backend.do_post_request(url, data=json.dumps(data, ignore_nan=True))
    _logger.info(f"Delivery response: {response_json}")
    job_id: str = response_json["jobId"]
    return job_id


def _process_file_paths(delivery_configs: List[DeliveryConfig]) -> Dict:
    """Get list of unique files/file paths to upload in S3"""
    processed_file_paths = {}
    for config in delivery_configs:
        file_path = config.file_path
        if file_path not in processed_file_paths:
            # create output filename to upload in s3
            output_file_name = _create_output_file_name(file_path)
            processed_file_paths[file_path] = output_file_name
        else:
            output_file_name = processed_file_paths[file_path]
        config.file_path = output_file_name
    return processed_file_paths


@_decorator.typechecked
def get_delivery_profile() -> Dict:
    """Fetch ftp delivery profiles using MDS API
    Returns:
        :obj:`dict`: FTP delivery profiles for the user.
    """
    url = f"{_get_mds_base_url()}v1/delivery-profile"
    _logger.info(f"Calling MDS /delivery-profile endpoint. ")
    mds_delivery_profiles: Dict[Any, Any] = _api_backend.do_get_request(url)
    _logger.info(f"MDS delivery profile response: {mds_delivery_profiles}")
    ftp_delivery_profiles = {"ftps": mds_delivery_profiles.get("ftps", [])}
    return ftp_delivery_profiles


@_decorator.typechecked
def delivery_status(job_id: str, poll: bool = True) -> str:
    """Polls the MDS delivery api with the job_id to check the delivery status.
    Args:
        job_id (:obj:`str`): The job id that we get from MDS endpoint
        poll (:obj:`bool`): Flag to poll the endpoint or not (default is set as True,
        as we want to poll the endpoint to get the status)

    Returns:
        :obj:`str`: Delivery status as given by the endpoint.
    """
    url = f"{_get_mds_base_url()}v1/jobs/{job_id}/status"
    _logger.info(f"Calling MDS /delivery/{job_id}/status endpoint")

    # polling the endpoint till it gets status response; raise exception error if occurred
    def _delivery_finished(res: dict) -> bool:
        try:
            return res["jobStatus"].lower() in ["done", "failed"]
        except KeyError as e:
            _logger.error(f"Could not find delivery status. Exception as {e}")
            raise

    if poll:
        _logger.info(f"Polling the request to get the delivery status.")
        try:
            polling2.poll(
                lambda: _api_backend.do_get_request(url),
                step=5,
                timeout=60,
                check_success=lambda response: _delivery_finished(response),
            )
        except polling2.TimeoutException as te:
            _logger.info(f"Timeout for the request. Exception as {te}.")

    response_json = _api_backend.do_get_request(url)
    _logger.info(f"Delivery status: {response_json}")

    return str(response_json["jobStatus"])  # "DONE" or "FAILED" or "RUNNING"


@_decorator.typechecked
def deliver(
    config: Union[DeliveryConfig, List[DeliveryConfig]],
    wait_for_delivery: bool = False,
) -> dict:
    """
    :bdg-ref-danger:`Upcoming Feature <../upcoming_feature>`

    Delivers a file from a notebook to an email or FTP.

    Args:
        config (:obj:`md.utils.DeliveryConfig`): An object that holds the delivery configuration information.

            All delivery methods (email, FTP, Delivery Profile) contains
                * file_path (:obj: `str`, `required`): the path to the file that will be delivered, including file name and extension.
                * delivered_file_name (:obj: `str`, `optional`): parameter for the user to specify the file name to be used when
                    delivered. No file extension should be provided. If it's omitted, the file_path value will use as default filename.
                *  include_timestamp (:obj: `bool`, `optional`): optional parameter for the user append a timestamp to the file name. Default value will be False. The format will be YYYYMMDDHHMMSS.

            For an email, it contains
                * method (:obj:`DeliveryType`): ``DeliveryType.EMAIL`` in this case.
                * email_address (:obj:`str`): email_address to send the file to.
            For FTP, it contains
                * method (:obj:`DeliveryType`): ``DeliveryType.FTP`` in this case.
                * user_name (:obj:`str`): the user_name to login to the ftp server.
                * password (:obj:`str`): the password to login to the ftp server.
                * server (:obj:`str`): the ftp server to use.
                * folder (:obj:`str`): the folder on the ftp server to upload the file to.
            For a Delivery Profile, it contains
                * method (:obj:`DeliveryType`): ``DeliveryType.DELIVERY_PROFILE`` in this case.
                * delivery_profile_id (:obj:`str`): delivery_profile_id that was setup by MDS. Can be retrieved with get_delivery_profile()
        wait_for_delivery (:obj:`bool`): Flag to poll the api to check the delivery status or not. (True=show status)

    :Returns:
        :obj:`dict`: A dictionary with the 'job_id', 'message', 'delivery_status' keys containing information about the delivery status

    :Examples:

    Deliver to an email address.

    ::

        import morningstar_data as md
        import pandas as pd

        df = pd.DataFrame({'a':[1], 'b':[2]})
        df.to_csv("test_export.csv")

        # Email example
            * For single file
                delivery_config = md.utils.DeliveryConfig(method=md.utils.DeliveryType.EMAIL, email_address="test@email.com", file_path="test_export.csv", delivered_file_name="Delivered_File_123", include_timestamp=True)
                md.utils.deliver(config=delivery_config)

            * For multiple files
                df = pd.DataFrame({'a':[1], 'b':[2]})
                df.to_csv("test_export2.csv")

                delivery_config1 = md.utils.DeliveryConfig(file_path="test_export.csv", method=md.utils.DeliveryType.EMAIL, email_address="test1@email.com", delivered_file_name="Delivered_File_123", include_timestamp=True)
                delivery_config2 = md.utils.DeliveryConfig(file_path="test_export2.csv", method=md.utils.DeliveryType.EMAIL, email_address="test2@email.com", delivered_file_name="Delivered_File_123", include_timestamp=True)
                md.utils.deliver([delivery_config1, delivery_config2])
    Deliver to a ftp server.

    ::

        import morningstar_data as md
        import pandas as pd

        df = pd.DataFrame({'a':[1], 'b':[2]})
        df.to_csv("test_export.csv")

        # FTP example
        delivery_config = md.utils.DeliveryConfig(method=md.utils.DeliveryType.FTP, user_name="test", password="test", server="ts.ftp.com", folder="data/", file_path="test_export.csv", delivered_file_name="Delivered_File_123")
        md.utils.deliver(config=delivery_config)

    Deliver to a Delivery Profile

    ::

        import morningstar_data as md
        import pandas as pd

        df = pd.DataFrame({'a':[1], 'b':[2]})
        df.to_csv("test_export.csv")

        # Delivery Profile example
        delivery_config = md.utils.DeliveryConfig(method=md.utils.DeliveryType.DELIVERY_PROFILE, delivery_profile_id="1234", file_path="test_export.csv", delivered_file_name="Delivered_File_123")
        md.utils.deliver(config=delivery_config)

    Multiple delivery within a single notebook.

    ::

        import morningstar_data as md
        import pandas as pd

        df = pd.DataFrame({'a':[1], 'b':[2]})
        df.to_csv("test_export.csv")

        ftp_delivery_config = md.utils.DeliveryConfig(file_path="test_export.csv", method=md.utils.DeliveryType.FTP, user_name="test", password="test", server="ts.ftp.com", folder="data/", delivered_file_name="Delivered_File_123")
        email_delivery_config = md.utils.DeliveryConfig(file_path="test_export.csv", method=md.utils.DeliveryType.EMAIL, email_address="test@email.com", delivered_file_name="Delivered_File_123", include_timestamp=False)
        md.utils.deliver([ftp_delivery_config, email_delivery_config])

    Multiple delivery for multiple files.

    ::

        import morningstar_data as md
        import pandas as pd

        df = pd.DataFrame({'a':[1], 'b':[2]})
        df.to_csv("test_export.csv")

        df = pd.DataFrame({'a':[1], 'b':[2]})
        df.to_csv("test_export2.csv")

        ftp_delivery_config = md.utils.DeliveryConfig(file_path="test_export.csv", method=md.utils.DeliveryType.FTP, user_name="test", password="test", server="ts.ftp.com", folder="data/", delivered_file_name="Delivered_File_123", include_timestamp=False)
        email_delivery_config = md.utils.DeliveryConfig(file_path="test_export2.csv", method=md.utils.DeliveryType.EMAIL, email_address="test@email.com", delivered_file_name="Delivered_File_123", include_timestamp=True)
        md.utils.deliver([ftp_delivery_config, email_delivery_config])

    Deliver with a specific filename.

    ::

        import morningstar_data as md
        import pandas as pd

        df = pd.DataFrame({'a':[1], 'b':[2]})
        df.to_csv("test_export.csv")

        # Email example
        delivery_config = md.utils.DeliveryConfig(method=md.utils.DeliveryType.EMAIL, email_address="test@email.com", file_path="test_export.csv", delivered_file_name="Delivered_Dataframe_123", include_timestamp=True)
        md.utils.deliver(config=delivery_config)
        # As we set include_timestamp to True, the file delivered will be named "Delivered_Dataframe_123_2024022819184484.csv"

        delivery_config = md.utils.DeliveryConfig(method=md.utils.DeliveryType.EMAIL, email_address="test@email.com", file_path="test_export.csv", delivered_file_name="Delivered_Dataframe_123")
        md.utils.deliver(config=delivery_config)
        # If include_timestamp set to False or if omitted, the file delivered will be named "Delivered_Dataframe_123.csv"

    Deliver without the delivered_file_name property

    ::

        import morningstar_data as md
        import pandas as pd

        df = pd.DataFrame({'a':[1], 'b':[2]})
        df.to_csv("test_export.csv")

        delivery_config = md.utils.DeliveryConfig(method=md.utils.DeliveryType.EMAIL, email_address="test@email.com", file_path="test_export.csv", include_timestamp=True)
        md.utils.deliver(config=delivery_config)
        # In this example we omitted the delivered_file_name property, so the file delivered will use the file_path and be named "test_export_2024022819184484.csv"


    :Errors:
        AccessDeniedError: Raised when the user is not authenticated.

        UnauthorizedDeliveryException: Raised when the user does not have permissions to deliver artifacts.

        BadRequestError: Raised when the user does not provide a properly formatted request.

        InternalServerError: Raised when the server encounters an unhandled error.

        NetworkExceptionError: Raised when the request fails to reach the server due to a network error.

        FileNotFoundError: Raised when the file path specified to be delivered does not exist
    """
    # Check for a feed_id, create one if not specified
    try:
        feed_id = int(os.environ["FEED_ID"])
    except KeyError:
        feed_id = _create_feed_id()

    # if single config is passed and it is not in list, convert into list
    if not isinstance(config, list):
        config = [config]

    # get list of unique files from delivery configs
    processed_file_paths = _process_file_paths(config)

    # upload files to s3
    for file_path, output_file_name in processed_file_paths.items():
        _upload_to_s3(file_path, output_file_name)

    _logger.info(f"Final config for creating post body: {config}")
    job_id: str = _mds_deliver_file(config, feed_id)
    res = {"job_id": job_id, "message": "Delivery has been submitted."}

    # if flag is set True, poll the api, return the delivery status
    res["delivery_status"] = delivery_status(job_id, poll=wait_for_delivery)
    if str(res["delivery_status"]).lower() == "failed":
        raise FailedDeliveryError((FAILED_DELIVERY_ERROR).format(res["job_id"])) from None
    return res



import warnings
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Union

from pandas import DataFrame

from .. import mdapi
from .._utils import (
    _get_data_points_total_columns,
    _get_user_cells_quota,
    data_point_dataframe_to_list,
)
from ..mdapi import RequestObject
from . import _decorator
from ._data_objects import DataPoints, Investments
from ._data_type import DryRunResults
from .data_type import TimeSeriesFormat


@dataclass
class InvestmentDataRequest(RequestObject):
    investments: Union[List[str], str, Dict[str, Any]]
    datapoints: Optional[Union[List[Dict[str, Any]], str, List[Any]]] = None
    time_series_format: TimeSeriesFormat = TimeSeriesFormat.WIDE
    display_name: Optional[bool] = False


@_decorator.typechecked
def investment_data(
    investments: Union[List[str], str, Dict[str, Any]],
    datapoints: Union[List[Dict[str, Any]], str, DataFrame],
) -> DataFrame:
    warnings.warn(
        "The investment_data function is deprecated and will be removed in the next major version. Use get_investment_data instead",
        FutureWarning,
        stacklevel=2,
    )
    return _get_investment_data(investments, datapoints, time_series_format=TimeSeriesFormat.WIDE)


# FIXME: In https://msjira.morningstar.com/browse/AL2-92, the error_handler decorator was removed from some methods
# to trigger QueryLimitException.
# There are multiple methods where the decorator was removed to trigger certain exceptions back to the caller:
#   - save_investment_list
#   - investment_data
#   - holding_dates
#   - asset_flow
# For the time being, we will leave this as-is to keep things working. Moving forward, we must re-design the
# exception handling so that we don't have these types of special cases where error_handler is only allowed on some
# methods but not others.
@_decorator.typechecked
def get_investment_data(
    investments: Union[List[str], str, Dict[str, Any]],
    data_points: Optional[Union[List[Dict[str, Any]], str, DataFrame, List[Any]]] = None,
    display_name: bool = False,
    dry_run: Optional[bool] = False,
    time_series_format: TimeSeriesFormat = TimeSeriesFormat.WIDE,
) -> Union[DataFrame, DryRunResults]:
    """Retrieve data for the specified investments and data points.

    Args:
        investments (:obj:`Union`, `required`): Defines the investments to fetch. Input can be:

            * Investment IDs (:obj:`list`, `optional`): Investment identifiers, in the format of SecId;Universe or just SecId. E.g., ["F00000YOOK;FO","FOUSA00CFV;FO"] or ["F00000YOOK","FOUSA00CFV"]. Use the `investments <./lookup.html#morningstar_data.direct.investments>`_ function to discover identifiers.
            * Investment List ID (:obj:`str`, `optional`): Saved investment list in Morningstar Direct. Currently, this function does not support lists that combine investments and user-created portfolios. Use the `get_investment_lists <./lists.html#morningstar_data.direct.user_items.get_investment_lists>`_ function to discover saved lists.
            * Search Criteria  ID (:obj:`str`, `optional`): Saved search criteria in Morningstar Direct. Use the `get_search_criteria <./search_criteria.html#morningstar_data.direct.user_items.get_search_criteria>`_ function to discover saved search criteria.
            * Search Criteria Condition (:obj:`dict`, `optional`): Search criteria definition. See details in the Reference section below or use the `get_search_criteria_conditions <./search_criteria.html#morningstar_data.direct.user_items.get_search_criteria_conditions>`_ function to discover the definition of a saved search criteria.

        data_points (:obj:`Union`, `optional`): Defines the data points to fetch. If not provided and investments are specified with a list ID or search criteria ID, the corresponding bound dataset will be used.

            * Data Point IDs (:obj:`List[Dict]`, `optional`): A list of dictionaries, each defining a data point and its (optional) associated settings.
            * Data Set ID (:obj:`str`, `optional`): Morningstar data set or user-created data set saved in Morningstar Direct. Use the `get_data_sets <./data_set.html#morningstar_data.direct.user_items.get_data_sets>`_ or `get_morningstar_data_sets <./data_set.html#morningstar_data.direct.get_morningstar_data_sets>`_ functions to discover saved data sets.
            * Data Point Settings (:obj:`DataFrame`, `optional`): A DataFrame of data point identifiers and their associated settings. Use the `get_data_set_details <./data_set.html#morningstar_data.direct.user_items.get_data_set_details>`_ function to discover data point settings from a saved data set.

        display_name (:obj:`bool`, `optional`): When true, the returned column names will match display names saved in the data set. Default is false.
        time_series_format (:obj:`md.direct.data_type.TimeSeriesFormat`, `optional`): Specifies the format of the time series data. Default is WIDE. This parameter accepts enumerations from `md.direct.data_type.TimeSeriesFormat`, which include:

            * WIDE: Data is presented in a wide format, where each row contains a different investment with values for each point in time in a different column.
            * LONG: Data is presented in a long format, where each row represents a single observation for a variable at a point in time.
            * LONG_WITHOUT_NAME: Similar to the long format but excludes the 'Name' column.

        dry_run(:obj:`bool`, `optional`): When True, the query will not be executed. Instead, a DryRunResults object will be returned with details about the query's impact on daily cell limit usage.

    :Returns:

        There are two return types:

        * DataFrame: A DataFrame object with investment data

        * DryRunResults: Is returned if dry_run=True is passed

          * estimated_cells_used: Number of cells by this query
          * daily_cells_remaining_before: How many cells are remaining in your daily cell limit before running this query
          * daily_cells_remaining_after: How many cells would be remaining in your daily cell limit after running this query
          * daily_cell_limit: Your total daily cell limit

    :Reference:

        Constructing a Search Criteria Condition dictionary:

        For example::

                    SEARCH_CRITERIA_CONDITION = {
                            "universeId": "cz",
                            "subUniverseId": "",
                            "subUniverseName": "",
                            "securityStatus": "activeonly",
                            "useDefinedPrimary": False,
                            "criteria": [
                                {"relation": "", "field": "HU338", "operator": "=", "value": "1"},
                                {"relation": "AND", "field": "HU863", "operator": "=", "value": "1"}
                            ]
                        }

        * universeId (:obj:`string`, `required`): Universe code (e.g., "FO"). Use the `get_investment_universes <./lookup.html#morningstar_data.direct.lookup.get_investment_universes>`_ function to explore available values or `download <../_static/assets/SearchCriteriaInfo.xlsx>`_ the reference file.
        * subUniverseId (:obj:`string`, `optional`): Sub-universe code, if applicable. See the refererece file above for available values.
        * subUniverseName (:obj:`string`, `optional`): Name of sub-universe, if applicable. See the refererece file above for available values.
        * securityStatus (:obj:`string`, `optional`): Security status, can be 'activeonly' or 'activeinactive'.
        * useDefinedPrimary (:obj:`Boolean`, `optional`): If set to true, Morningstar Direct user-defined settings will be used.
        * criteria (:obj:`List[Dict[]]`, `required`): Custom search conditions. Dictionary fields described below.

        The 'criteria' object consists of the following fields:
            * field (:obj:`string`): Data point identifier
            * value (:obj:`string`): Value to compare against the data point.
            * criteria (:obj:`List[Dict[]]`): A nested list of additional custom search conditions.
            * relation (:obj:`string`, `required`): Boolean condition used when joining with the previous condition. Accepts an empty string (for the first condition, as no previous condition exists to join with), 'Or', or 'And'.
            * operator (:obj:`string`): Operator used to compare field value to value. Possible operators include '=', '!=', '<', '>', '<=', '>=', 'like' (data contains value), and 'not like' (data does not contain value).

            For example::

                [
                    {
                        "field": "eb919bcc-c097-4fe3-898c-470d8b89dde1"
                        "operator": "="
                        "relation": ""
                        "value": "122"
                    },
                    {
                        "criteria": [
                            {...Condition1 },
                            {...Condition2 },
                            {...Condition3 },
                        ],
                        "relation": "And"
                    },
                    {
                        "field": "LS466"
                        "operator": "="
                        "relation": "Or"
                        "value": "FH"
                    }
                ]

    :Examples:

    Get investment data for the given investments and data points.
    ::

        import morningstar_data as md

        df = md.direct.get_investment_data(
            investments=["F0AUS05U7H", "F000010NJ5"],
            data_points=[
                {"datapointId": "OS01W"}, # Name
                {"datapointId": "LS05M"}, # Base Currency
                {
                    "datapointId": "HP010", # Monthly Return
                    "isTsdp": True,
                    "currency": "CNY",
                    "startDate": "2021-03-01",
                    "endDate": "2021-08-31",
                },
            ],
        )
        df

    :Output:
        =================  ===================================  ================= =========================  =========================
        Id                 Name                                 Base Currency     Monthly Return 2021-03-31  Monthly Return 2021-04-30
        =================  ===================================  ================= =========================  =========================
        F0AUS05U7H         Walter Scott Global Equity           Australian Dollar                  3.726094                   3.078352
        F000010NJ5         Vontobel Emerging Markets Eq U1 USD  US Dollar                         -0.417526                  -0.376890
        =================  ===================================  ================= =========================  =========================


        Get investment data for a saved investment list and data points.

        ::

            import morningstar_data as md

            df = md.direct.get_investment_data(
                investments="a727113a-9557-4378-924f-5d2ba553f687", # Replace with a valid List ID
                data_points=[{"datapointId": "HS793", "isTsdp": True}],
            )
            df


        ==============  =================================  =============================  =============================  =============================
        Id              Name                               Daily Return Index 2021-09-23  Daily Return Index 2021-09-24  Daily Return Index 2021-09-25
        ==============  =================================  =============================  =============================  =============================
        FOUSA00DFS;FO   BlackRock Global Allocation Inv A  129.92672                      129.56781                      129.56781
        ==============  =================================  =============================  =============================  =============================

        Get investment data for a saved search criteria and data points

        ::

            import morningstar_data as md

            df = md.direct.get_investment_data(
                investments="4216254", # Replace with a valid Search Criteria ID
                data_points=[{"datapointId": "12", "isTsdp": True}]
            )
            df


        ==============  =======================  =============================
        Id              Name                     Beta 2018-10-01 to 2021-09-30
        ==============  =======================  =============================
        FOUSA06JNH;FO   DWS RREEF Real Assets A    0.654343
        ==============  =======================  =============================

        Get investment data for a custom search criteria and data points.

        ::

            import morningstar_data as md

            SEARCH_CRITERIA_CONDITION = {
                "universeId": "cz",
                "subUniverseId": "",
                "subUniverseName": "",
                "securityStatus": "activeonly",
                "useDefinedPrimary": False,
                "criteria": [
                    {"relation": "", "field": "HU338", "operator": "=", "value": "1"},
                    {"relation": "AND", "field": "HU863", "operator": "=", "value": "1"},
                ],
            }

            df = md.direct.get_investment_data(
                investments=SEARCH_CRITERIA_CONDITION,
                data_points=[{"datapointId": "HS793", "isTsdp": True}],
            )
            df


        =  ===============  =====================================  =============================  ===  =============================
        #  Id               Name                                   Daily Return Index 2022-02-18  ...  Daily Return Index 2022-03-17
        =  ===============  =====================================  =============================  ===  =============================
        0  FOUSA06UOR;CZ    Columbia Trust Stable Government Fund  None                           ...  None
        1  FOUSA06UWL;CZ    Columbia Trust Stable Income Fund      88.8333                        ...  90.7781
        =  ===============  =====================================  =============================  ===  =============================

    Get investment data for the given investments and data points in the long format.
    ::

            import morningstar_data as md

        df = md.direct.get_investment_data(
            investments=["F0AUS05U7H", "F000010NJ5"],
            data_points=[
                {"datapointId": "OS01W"}, # Name
                {"datapointId": "LS05M"}, # Base Currency
                {
                    "datapointId": "HP010", # Monthly Return
                    "isTsdp": True,
                    "currency": "CNY",
                    "startDate": "2021-03-01",
                    "endDate": "2021-08-31",
                },
            ],
            time_series_format=md.direct.data_type.TimeSeriesFormat.LONG
        )
        df


        =================  ===================================  ================= =========================  =========================
        Id                 Name                                 Base Currency     Date                       Monthly Return
        =================  ===================================  ================= =========================  =========================
        F0AUS05U7H         Walter Scott Global Equity           Australian Dollar                2021-04-30                   3.726094
        F0AUS05U7H         Walter Scott Global Equity           Australian Dollar                2021-05-31                   3.078352
        F0AUS05U7H         Walter Scott Global Equity           Australian Dollar                2021-06-30                   3.078352
        F000010NJ5         Vontobel Emerging Markets Eq U1 USD  US Dollar                        2021-04-30                  -0.417526
        F000010NJ5         Vontobel Emerging Markets Eq U1 USD  US Dollar                        2021-05-31                  -0.376890
        F000010NJ5         Vontobel Emerging Markets Eq U1 USD  US Dollar                        2021-06-30                  3.078352
        =================  ===================================  ================= =========================  =========================


    """

    result = _get_investment_data(
        investments=investments,
        data_points=data_points,
        display_name=display_name,
        time_series_format=time_series_format,
        dry_run=dry_run,
    )
    return result


@_decorator.typechecked
def _get_investment_data(
    investments: Union[List[str], str, Dict[str, Any]],
    data_points: Optional[Union[List[Dict[str, Any]], str, DataFrame, List[Any]]] = None,
    display_name: bool = False,
    dry_run: Optional[bool] = False,
    time_series_format: TimeSeriesFormat = TimeSeriesFormat.WIDE,
) -> Union[DataFrame, DryRunResults]:
    if dry_run:
        investment_param = Investments(investments)
        data_point_param = DataPoints(data_points)
        investment_id_list = investment_param.get_investment_ids()
        data_point_settings = data_point_param.get_data_points()
        data_point_settings = data_point_settings.where(data_point_settings.notnull(), None)
        data_point_list = data_point_settings.to_dict(orient="records")
        estimated_cells_used = _get_data_points_total_columns(data_point_list) * len(investment_id_list)
        user_cells_quota = _get_user_cells_quota()
        dry_run_results = DryRunResults(
            estimated_cells_used=estimated_cells_used,
            daily_cells_remaining_before=user_cells_quota["daily_cell_remaining"],
            daily_cell_limit=user_cells_quota["daily_cell_limit"],
        )
        return dry_run_results

    if isinstance(data_points, DataFrame):
        data_points = data_point_dataframe_to_list(data_points)

    return mdapi.call_remote_function(
        "get_investment_data",
        InvestmentDataRequest(
            investments=investments, datapoints=data_points, time_series_format=time_series_format, display_name=display_name
        ),
    )



import dataclasses
import os
import re
from enum import Enum
from pathlib import Path

from ..direct._exceptions import BadRequestException, ValueErrorException


class DeliveryConfigErrorMessage(Enum):
    EMAIL = """
        Invalid Email Address.
        Example:
        - email_address = "analyticslab.user@morningstar.com"
    """
    FTP = """
        Invalid FTP configuration.
        The username and password must not be empty
        The server must not include any protocal, like "ftp://".
        The folder must not include a root path '/'.
        Example:
        - user_name = "analyticslab_user"
        - password = "password123"
        - server = "ftp.morningstar.com"
        - folder = "folder_1/folder_2/"
    """
    DELIVERY_PROFILE = """
        Invalid delivery profile Id.
        The delivery profile id must not be empty
        The delivery profile id must be a digit string
        Example:
        - delivery_profile_id = "12345"
    """


class DeliveryType(Enum):
    EMAIL = 1
    FTP = 2
    DELIVERY_PROFILE = 3


@dataclasses.dataclass
class DeliveryConfig:
    """Container for delivery configuration settings
    Args:
        method (:obj:`DeliveryType`): Required, the type of delivery, see DeliveryType for supported options
        email_address (:obj:`str`): a valid email_address to send the file to, needed if using "email" delivery option
        user_name (:obj:`str`): the user_name to login to the ftp server, needed if using "ftp" method
        password (:obj:`str`): the password to login to the ftp server, needed if using "ftp" method
        server (:obj:`str`): the ftp server to use, needed if using "ftp" method. The server address should not include "ftp://"
        folder (:obj:`str`): the folder on the ftp server to upload the file to, needed if using "ftp" method. The folder should not include a folder path like "/folder"
        file_path (:obj:`str`, `required`): the path to the file that will be delivered, including file name and extension.
        delivered_file_name (:obj:`str`, `optional`): the parameter for the user to specify the file name to be used when
            delivered. No file extension should be provided. If it's omitted, the file_path value will use as default filename. Valid characters are `., -, _, (, ), , a-z, A-Z, 0-9`. If the file
            name includes non-valid characters, an exception will be raised. If the property is omitted and the file_path is "data/folder/test_export.csv"
            the delivered_file_name will be "test_export.csv" or "test_export_2024022819184484.csv" based on the value of include_timestamp property.
        include_timestamp (:obj: `bool`, `optional`): optional parameter for the user append a timestamp to the file name. Default value will be False. The format will be YYYYMMDDHHMMSS.

    :Examples:

    Valid delivery config

    ::
        email_address = "test@morningstar.com"
        delivery_config_email = md.utils.DeliveryConfig(method=md.utils.DeliveryType.EMAIL, email_address=email_address, file_path="test_export.csv", delivered_file_name="Delivered_File_123")

        ftp_username = "test"
        ftp_password = "test_pwd"
        ftp_server = "ts.ftp.com"
        ftp_folder = "data/"
        delivery_config_ftp = md.utils.DeliveryConfig(method=md.utils.DeliveryType.FTP, user_name=ftp_username, password=ftp_password, server=ftp_server, folder=ftp_folder, file_path="test_export.csv", delivered_file_name="Delivered_File_123", include_timestamp=True)
        # As include_timestamp is set to True, the final file name will be Delivered_File_123_2024022819184484.csv

        delivery_profile_id = "12345"
        delivery_config_email = md.utils.DeliveryConfig(method=md.utils.DeliveryType.DELIVERY_PROFILE, delivery_profile_id=delivery_profile_id, file_path="test_export.csv", delivered_file_name="Delivered_File_123", include_timestamp=True)
        # As include_timestamp is set to True, the final file name will be Delivered_File_123_2024022819184484.csv

    Valid delivery config if doing multiple files or destination deliveries (Specify file_path and delivered_file_name)

    ::
        delivery_config_email = md.utils.DeliveryConfig(file_path="test_export.csv", method=md.utils.DeliveryType.EMAIL, email_address=email_address, delivered_file_name="Delivered_Dataframe_123", include_timestamp=False)

    Deliver without the delivered_file_name property

    ::

        import morningstar_data as md
        import pandas as pd

        df = pd.DataFrame({'a':[1], 'b':[2]})
        df.to_csv("test_export.csv")

        delivery_config = md.utils.DeliveryConfig(method=md.utils.DeliveryType.EMAIL, email_address="test@email.com", file_path="test_file.csv", include_timestamp=True)
        md.utils.deliver(config=delivery_config)
        # In this example we omitted the delivered_file_name property, so the file delivered will use the file_path and be named "test_file_2024022819184484.csv"

        delivery_config = md.utils.DeliveryConfig(method=md.utils.DeliveryType.EMAIL, email_address="test@email.com", file_path="test_file.csv", include_timestamp=False)
        md.utils.deliver(config=delivery_config)
        # In this example we omitted the delivered_file_name property and the include_timestamp property is false,
        # so the file delivered will use the file_path and be named "test_file.csv"

    Invalid delivery config

    ::
        email_address = "@invalid_email_address@morningstar.com"
        delivery_config_email = md.utils.DeliveryConfig(method=md.utils.DeliveryType.EMAIL, email_address=email_address, delivered_file_name="Delivered_Dataframe_123")
        >>> ValueErrorException: Invalid Email Address

        ftp_username = "test"
        ftp_password = "test_pwd"
        ftp_server = "ftp://ts.ftp.com"
        ftp_folder = "/data/"
        delivery_config_ftp = md.utils.DeliveryConfig(method=md.utils.DeliveryType.FTP, user_name=ftp_username, password=ftp_password, server=ftp_server, folder=ftp_folder, delivered_file_name="Delivered_Dataframe_123")
        >>> ValueErrorException: Invalid FTP configuration

        delivery_profile_id = "12ABC"
        delivery_config_email = md.utils.DeliveryConfig(method=md.utils.DeliveryType.DELIVERY_PROFILE, delivery_profile_id=delivery_profile_id, delivered_file_name="Delivered_Dataframe_123", include_timestamp=True)
        >>> ValueErrorException: Invalid delivery profile Id

        delivery_profile_id = "12ABC"
        delivery_config_email = md.utils.DeliveryConfig(method=md.utils.DeliveryType.DELIVERY_PROFILE, delivery_profile_id=delivery_profile_id)
        >>> BadRequestException: Delivered File Name can not be empty. Please provide a valid file name.
    """

    method: DeliveryType
    delivered_file_name: str = ""
    email_address: str = ""
    delivery_profile_id: str = ""
    user_name: str = ""
    password: str = ""
    server: str = ""
    folder: str = ""
    file_path: str = ""
    include_timestamp: bool = False

    def __post_init__(self) -> None:
        if not self.is_delivery_type_valid():
            raise ValueErrorException(DeliveryConfigErrorMessage[self.method.name].value)

        if not self._is_file_path_valid():
            error_message = (
                f"The specified file '{self.file_path}' does not exist or the path is invalid. Please provide a valid file path."
            )
            raise BadRequestException(error_message)

        if not self.delivered_file_name:
            self.delivered_file_name = Path(self.file_path).stem

        invalid_file_name = self._is_file_name_valid(self.delivered_file_name)
        if invalid_file_name:
            error_message = f"delivered_file_name: {self.delivered_file_name} is invalid. The following can't be used in the file name: '{invalid_file_name}'"
            raise BadRequestException(error_message)

    def _is_email_valid(self) -> bool:
        """Check whether the email config provided is valid"""
        email_regex = re.compile(
            r"""\b             # a word boundary
                [a-z0-9._%+-]+ # the email prefix part, any character in the list
                @              # the @ part
                [a-z0-9.-]+    # the email domain name part, any character in the list
                \.             # the dot part
                [a-z]{2,}      # the top-level domain part, any alphabetical character and length >= 2
                \b             # a word boundary
            """,
            re.X,
        )
        return re.fullmatch(email_regex, self.email_address.lower()) is not None

    def _is_delivery_profile_valid(self) -> bool:
        """Checks if the provided delivery profile ID is valid.
        At present, the MDS delivery endpoint allows any user to deliver to any profile ID.
        In the future, MDS need to add validation based on the user's company.
        """
        return self.delivery_profile_id.isdigit()

    def _is_file_name_valid(self, file_name: str) -> list:
        """Checks if the file name used for the delivered file is valid"""
        # A filename can include 'a-z', 'A-Z', '0-9', '-', '_', '(', ')', ' ', '.'
        invalid_chars_regex = re.compile(r"[^a-zA-Z\d\-_\(\)\ .]")
        return invalid_chars_regex.findall(file_name)

    def _is_ftp_valid(self) -> bool:
        """Check whether the ftp config provided is valid"""
        if self.user_name == "" or self.password == "":
            return False

        server_regex = re.compile(
            r"""\b          # a word boundary
                [a-z0-9.-]+ # the domain name part, any character in the list
                \.          # the dot part
                [a-z]{2,}   # the top-level domain part, any alphabetical character and length >= 2
                \b          # a word boundary
            """,
            re.X,
        )
        if not re.fullmatch(server_regex, self.server.lower()):
            return False

        folder_regex = re.compile(
            r"""
                ^[^\/]                   # cannot start with /
                (
                    [\w\d\s\._%+-\[\]]+  # any character in the list
                    [\/]{1,1}            # ending with /
                )+                       # capturing group
            """,
            re.X,
        )
        if not re.fullmatch(folder_regex, self.folder):
            return False

        return True

    def _is_file_path_valid(self) -> bool:
        """Check whether the file path provided is valid"""
        return os.path.isfile(self.file_path)

    def is_delivery_type_valid(self) -> bool:
        is_email_valid = self.method == DeliveryType.EMAIL and self._is_email_valid()
        is_ftp_valid = self.method == DeliveryType.FTP and self._is_ftp_valid()
        is_delivery_profile_valid = self.method == DeliveryType.DELIVERY_PROFILE and self._is_delivery_profile_valid()
        return is_email_valid or is_ftp_valid or is_delivery_profile_valid



from dataclasses import dataclass, field
from enum import Enum
from typing import List, Optional

DEFAULT_POLL_INTERVAL_MILLISECONDS = 1000
DEFAULT_POLL_TIMEOUT_SECONDS = 900  # seconds (15 minutes)


class TaskStatus(str, Enum):
    SUCCESS = "success"
    PENDING = "pending"
    FAILURE = "failure"


@dataclass
class TaskResult:
    dataframe_file_url: str
    columns_with_list_values: List[str] = field(default_factory=list)


@dataclass
class MdapiTask:
    id: str
    # TODO: Add request_id
    poll_url: str
    status: TaskStatus = TaskStatus.PENDING
    result: Optional[TaskResult] = None
    error_class: Optional[str] = None
    error_message: Optional[str] = None
    next_poll_delay_milliseconds: int = DEFAULT_POLL_INTERVAL_MILLISECONDS
    task_timeout_seconds: int = DEFAULT_POLL_TIMEOUT_SECONDS

    def is_complete(self) -> bool:
        return self.status != TaskStatus.PENDING

    def is_successful(self) -> bool:
        return self.status == TaskStatus.SUCCESS

    def failed(self) -> bool:
        return self.status == TaskStatus.FAILURE

    def __post_init__(self) -> None:
        if isinstance(self.result, dict):
            # This is initted from a API response json object with MdapiTask(**response.json()), so the initial
            # init will be called with a dict, even though we are specifying a TaskResult. Tell mypy to not warn about this.
            self.result = TaskResult(**self.result)  # type: ignore


@dataclass
class RequestObject:
    pass



import time
import uuid
import warnings
from dataclasses import dataclass
from datetime import datetime
from typing import Any, Dict, List, Optional, Union
from urllib.parse import quote

import pandas as pd
import simplejson as json
from pandas import DataFrame

from .. import datalake, mdapi
from .._base import _logger
from .._utils import _get_user_cells_quota
from ..mdapi import RequestObject
from . import _decorator, _utils
from . import investment as inv
from . import portfolio
from ._backend_apis._holdings_backend import (
    AMSAPIBackend,
    FoFAPIBackend,
    HoldingAPIBackend,
)
from ._config import _Config
from ._config_key import FORMAT_DATE
from ._data_objects import Investments
from ._data_type import DryRunResults
from ._error_messages import (
    BAD_REQUEST_ERROR_INCLUDE_ALL_DATE,
    BAD_REQUEST_ERROR_INVALID_LOOKTHROUGH_HOLDING_TYPE,
    BAD_REQUEST_ERROR_INVALID_PORTFOLIO_ID,
    BAD_REQUEST_ERROR_NO_INVESTMENT_IDS,
    BAD_REQUEST_ERROR_NO_PORTFOLIO_DATA,
    BAD_REQUEST_ERROR_NO_START_DATE,
)
from ._exceptions import BadRequestException

_config = _Config()

_holding_api_request = HoldingAPIBackend()
_fof_api_request = FoFAPIBackend()
_ams_api_request = AMSAPIBackend()

MASTER_PORTFOLIO_ID = "MasterPortfolio Id"


@dataclass
class HoldingsRequest(RequestObject):
    investments: Union[List[str], str, Dict[str, Any]]
    date: Optional[str]
    start_date: Optional[str]
    end_date: Optional[str]


@_decorator.typechecked
def holdings(
    investment_ids: List[str],
    date: Optional[str] = None,
    start_date: Optional[str] = None,
    end_date: Optional[str] = None,
) -> DataFrame:
    warnings.warn(
        "The holdings function is deprecated and will be removed in the next major version. Use get_holdings instead",
        FutureWarning,
        stacklevel=2,
    )
    return get_holdings(investment_ids, date, start_date, end_date)


@_decorator.typechecked
def get_holdings(
    investments: Optional[Union[List[str], str, Dict[str, Any]]] = None,  # Make it required once investment_ids get removed,
    date: Optional[str] = None,
    start_date: Optional[str] = None,
    end_date: Optional[str] = None,
    investment_ids: Optional[List[str]] = None,
    dry_run: Optional[bool] = False,
) -> Union[DataFrame, DryRunResults]:
    """Returns holdings for the specified investments and date or date range. If the date is not specified, the function uses the latest portfolio date by default.

    Args:
        investments (:obj:`Union`, `required`): Defines the investments to fetch. Input can be:

                * Investment IDs (:obj:`list`, `optional`): Investment identifiers, in the format of SecId;Universe or just SecId. E.g., ["F00000YOOK;FO","FOUSA00CFV;FO"] or ["F00000YOOK","FOUSA00CFV"]. Use the `investments <./lookup.html#morningstar_data.direct.investments>`_ function to discover identifiers.
                * Investment List ID (:obj:`str`, `optional`): Saved investment list in Morningstar Direct. Currently, this function does not support lists that combine investments and user-created portfolios. Use the `get_investment_lists <./lists.html#morningstar_data.direct.user_items.get_investment_lists>`_ function to discover saved lists.
                * Search Criteria  ID (:obj:`str`, `optional`): Saved search criteria in Morningstar Direct. Use the `get_search_criteria <./search_criteria.html#morningstar_data.direct.user_items.get_search_criteria>`_ function to discover saved search criteria.
                * Search Criteria Condition (:obj:`dict`, `optional`): Search criteria definition. See details in the Reference section below or use the `get_search_criteria_conditions <./search_criteria.html#morningstar_data.direct.user_items.get_search_criteria_conditions>`_ function to discover the definition of a saved search criteria.

        date (:obj:`str`, `optional`): The portfolio date for which to retrieve data. The format is YYYY-MM-DD.
            For example, "2020-01-01". If a date is provided, then the `start_date` and `end_date` parameters are ignored.
            An exception is thrown if `start_date` or `end_date` is provided along with `date`.
        start_date (:obj:`str`, `optional`): The start date for retrieving data. The format is
            YYYY-MM-DD. For example, "2020-01-01". An exception is thrown if `date` is provided along with `start_date`.
        end_date (:obj:`str`, `optional`): The end date for retrieving data. If no value is provided for
            `end_date`, current date will be used. The format is YYYY-MM-DD. For example, "2020-01-01". An exception is
            thrown if `date` is provided along with `end_date`.
        investment_ids (:obj:`list`): DEPRECATED, A list of investment IDs. The investment ID format is SecId;Universe or just SecId.
            For example: ["F00000YOOK;FO","FOUSA00CFV;FO"] or ["F00000YOOK","FOUSA00CFV"].
        dry_run(:obj:`bool`, `optional`): When True, the query will not be executed. Instead, a DryRunResults object will be returned with details about the query's impact on daily cell limit usage.
            Note: A few cells from your daily quota will be consumed in order to get the most recent number of holdings for each portfolio.

    :Returns:

        There are two return types:

        DataFrame: A DataFrame object with holdings data. DataFrame columns include:

        * investmentId
        * masterPortfolioId
        * portfolioDate
        * holdingId
        * bondId
        * name
        * secId
        * isin
        * cusip
        * weight
        * shares
        * marketValue
        * sharesChanged
        * currency
        * ticker
        * detailHoldingType

        DryRunResults: Is returned if dry_run=True is passed

        * estimated_cells_used: Number of cells by this query
        * daily_cells_remaining_before: How many cells are remaining in your daily cell limit before running this query
        * daily_cells_remaining_after: How many cells would be remaining in your daily cell limit after running this query
        * daily_cell_limit: Your total daily cell limit

    Raises:
        ValueErrorException: Raised when the `investments` parameter is invalid.

    :Examples:
        Retrieve holdings for investment "FOUSA00KZH" on "2020-12-31".

    ::

        import morningstar_data as md

        df = md.direct.get_holdings(investments=["FOUSA00KZH"], date="2020-12-31")
        df

    :Output:
        =============  =================  ===  ======  =================
        investmentId   masterPortfolioId  ...  ticker  detailHoldingType
        =============  =================  ===  ======  =================
        FOUSA00KZH     6079               ...  CBRE    EQUITY
        FOUSA00KZH     6079               ...  GOOGL   EQUITY
        ...
        =============  =================  ===  ======  =================

    """
    date = date or None
    start_date = start_date or None
    end_date = end_date or None

    if investment_ids is not None:
        warnings.warn(
            "The investment_ids argument is deprecated and will be removed in the next major version. Use investments instead",
            FutureWarning,
            stacklevel=2,
        )

    investments_object = investments or investment_ids

    if not investments_object:
        raise mdapi.BadRequestError("The `investments` parameter must be included when calling get_holdings") from None

    # TODO: Migrate to the API after initial v1 migration:
    if dry_run:
        investments_for_dry_run = Investments(investments or investment_ids)
        investment_id_list = investments_for_dry_run.get_investment_ids()
        investment_date_info = _get_dates(investment_id_list, date, start_date, end_date)
        return _get_dry_run_results(investment_date_info)

    return mdapi.call_remote_function(
        "get_holdings",
        HoldingsRequest(investments=investments_object, date=date, start_date=start_date, end_date=end_date),
    )


def _get_dates_by_investments(investment_ids: list) -> List[Any]:
    ids = quote(",".join(investment_ids), "utf-8")
    url = f"{_config.securitydata_service_url()}v2/securities/{ids}/portfolio-dates"
    result: List[Any] = _holding_api_request.do_get_request(url)
    return result


@_decorator.typechecked
def holding_dates(investment_ids: List[str]) -> DataFrame:
    warnings.warn(
        "The holding_dates function is deprecated and will be removed in the next major version. Use get_holding_dates instead",
        FutureWarning,
        stacklevel=2,
    )
    return get_holding_dates(investment_ids)


@_decorator.typechecked
def get_holding_dates(investment_ids: List[str]) -> DataFrame:
    """Returns all dates with available holdings data for the given investment.

    Args:
        investment_ids (:obj:`list`): A list of investment IDs. The investment ID format is SecId;Universe or just SecId.
            For example: ["F00000YOOK;FO","FOUSA00CFV;FO"] or ["F00000YOOK","FOUSA00CFV"].

    :Returns:
        DataFrame: A DataFrame object with portfolio date data. DataFrame columns include:

        * secId
        * masterPortfolioId
        * date
        * suppression
        * suppressionHoldingNumber

    :Examples:
        Retrieve portfolio dates for investment "FOUSA00KZH".

    ::

        import morningstar_data as md

        df = md.direct.get_holding_dates(investment_ids=["FOUSA06JNH"])
        df

    :Output:
        ==========  =================  ==========  ===========  ========================
        secId       masterPortfolioId  date        suppression  suppressionHoldingNumber
        ==========  =================  ==========  ===========  ========================
        FOUSA06JNH  210311             2021-08-31  False        None
        FOUSA06JNH  210311             2021-07-31  False        None
        ...
        ==========  =================  ==========  ===========  ========================

    """
    if not investment_ids:
        raise BadRequestException(BAD_REQUEST_ERROR_NO_INVESTMENT_IDS) from None
    response_json = _get_dates_by_investments(investment_ids)
    if isinstance(response_json, list) and response_json:
        details = []
        for date_info in response_json:
            sec_id = date_info.get("secId", "")
            master_portfolio_id = date_info.get("masterPortfolioId", "")
            details.extend(
                [
                    {
                        "secId": sec_id,
                        "masterPortfolioId": master_portfolio_id,
                        "date": portfolio_date.get("date", ""),
                        "suppression": portfolio_date.get("suppression", False),
                        "suppressionHoldingNumber": portfolio_date.get("suppressionHoldingNumber", None),
                    }
                    for portfolio_date in date_info.get("portfolioDates", [])
                ]
            )
        return DataFrame(details)
    else:
        return DataFrame()


def _get_dates(
    investment_ids: list,
    date: Optional[str] = None,
    start_date: Optional[str] = None,
    end_date: Optional[str] = None,
) -> dict:
    dates_raw: Dict[str, Optional[str]] = {
        "date": date,
        "start_date": start_date,
        "end_date": end_date,
    }
    dates: Dict[str, str] = {}

    for key, value in dates_raw.items():
        if value is not None and len(value.strip()) > 0:
            dates[key] = _utils._format_date(value)
        elif key == "end_date" and dates_raw["start_date"] is not None:
            dates[key] = _utils._format_date(time.strftime(FORMAT_DATE, time.localtime()))

    response_json = _get_dates_by_investments(investment_ids)
    if not isinstance(response_json, list) or not response_json:
        raise BadRequestException(BAD_REQUEST_ERROR_NO_PORTFOLIO_DATA) from None

    if not bool(dates):
        return _handle_no_date(response_json)
    elif "date" in dates:
        return _handle_single_date(response_json, **dates)

    return _handle_date_range(response_json, **dates)


def _validate_date(date: Optional[str], start_date: Optional[str], end_date: Optional[str]) -> None:
    if date:
        if start_date or end_date:
            raise BadRequestException(BAD_REQUEST_ERROR_INCLUDE_ALL_DATE) from None
    else:
        if not start_date and end_date:
            raise BadRequestException(BAD_REQUEST_ERROR_NO_START_DATE) from None


def _handle_single_date(date_info_list: list, date: str) -> dict:
    investment_date_info = dict()
    for date_info in date_info_list:
        available_date = _get_available_date(date, date_info)
        if available_date is not None:
            master_portfolio_id = date_info.get("masterPortfolioId", "")
            investment_date_info[date_info.get("secId", "")] = {
                "masterPortfolioId": master_portfolio_id,
                "portfolioDates": [available_date],
            }
    return investment_date_info


def _get_available_date(date: str, date_info: dict = dict()) -> Optional[str]:
    date_objs = date_info.get("portfolioDates", [])
    available_date = None
    for date_obj in date_objs:
        if date_obj.get("date", "") == date:
            available_date = date
            break
        elif date_obj.get("date", "") < date and available_date is None:
            available_date = date_obj.get("date", "")
    return available_date


def _handle_no_date(date_info_list: list) -> dict:
    investment_date_info = dict()
    for date_info in date_info_list:
        portfolio_dates = date_info.get("portfolioDates", [])
        if portfolio_dates:
            date = next(
                (x.get("date", "") for x in portfolio_dates if x.get("date", "") is not None and len(x.get("date", "")) > 0),
                None,
            )
            if date is not None and len(date) > 0:
                master_portfolio_id = date_info.get("masterPortfolioId", "")
                investment_date_info[date_info.get("secId", "")] = {
                    "masterPortfolioId": master_portfolio_id,
                    "portfolioDates": [date],
                }
    return investment_date_info


def _handle_date_range(date_info_list: list, start_date: str, end_date: str) -> dict:
    investment_date_info = dict()
    for date_info in date_info_list:
        sec_id = date_info.get("secId", "")
        master_portfolio_id = date_info.get("masterPortfolioId", "")
        portfolio_dates = date_info.get("portfolioDates", [])
        date_list = []
        for portfolio_date in portfolio_dates:
            single_date = portfolio_date.get("date", "")
            if start_date <= single_date <= end_date:
                date_list.append(single_date)
        if date_list:
            investment_date_info[sec_id] = {
                "masterPortfolioId": master_portfolio_id,
                "portfolioDates": date_list,
            }
    return investment_date_info


def _get_dry_run_results(investment_date_info: dict) -> DryRunResults:
    total_rows = 0
    data_points = [
        {"datapointId": "HS008", "isTsdp": False, "displayName": "long_all"},
        {"datapointId": "HS777", "isTsdp": False, "displayName": "short_other"},
        {"datapointId": "HS879", "isTsdp": False, "displayName": "short_stock"},
        {"datapointId": "HS880", "isTsdp": False, "displayName": "short_bonds"},
    ]

    if len(investment_date_info) > 0:
        n_holdings_per_investment_df = inv.get_investment_data(list(investment_date_info), data_points)
        n_holdings_per_investment_df = n_holdings_per_investment_df.fillna(0)

        for row_inv in n_holdings_per_investment_df.itertuples():
            this_inv_date_info = investment_date_info.get(row_inv.Id)
            if this_inv_date_info:
                n_dates = len(this_inv_date_info["portfolioDates"])
            n_holdings = row_inv.long_all + row_inv.short_other + row_inv.short_stock + row_inv.short_bonds
            total_rows = total_rows + (n_dates * int(n_holdings))

    total_columns = 16
    estimated_cells_used = total_columns * total_rows
    user_cells_quota = _get_user_cells_quota()
    dry_run_results = DryRunResults(
        estimated_cells_used=estimated_cells_used,
        daily_cells_remaining_before=user_cells_quota["daily_cell_remaining"],
        daily_cell_limit=user_cells_quota["daily_cell_limit"],
    )
    return dry_run_results


def _get_ids_from_lake_house(df_lookthrough: pd.DataFrame) -> pd.DataFrame:
    sql = """SELECT distinct
            a.investment_id
            ,a.isin
            ,a.cusip
            ,a.weight
            ,a.marketValue as market_value
            ,a.security_name
            ,a.currency
            ,a.morningstar_instrument_type_code
            ,b.entity_id
            ,b.morningstar_entity_name
            ,b.performance_id as primary_performance_id
            FROM lookthrough_temp a
                left join platform__reference__prd.investment_lookup b
                    on a.investment_id=b.investment_id and b.is_primary = True
            where (b.is_primary = True or b.investment_id is null)
    """
    df = datalake.query(sql, temp_tables=[datalake.TempTable("lookthrough_temp", df_lookthrough)])
    return df


def _get_fof_secid_and_investment_types(df_top_level_holdings: pd.DataFrame) -> pd.DataFrame:
    # create DataFrame with Direct secId and Direct securityType
    # df_top_level_holdings_expanded = df_top_level_holdings["HoldingId"].str.split(";", expand=True).rename(columns={0: "secid", 1: "sectype"})
    df_top_level_holdings[["fof_secid", "fof_sectype"]] = (
        df_top_level_holdings["HoldingId"].str.split(";", expand=True).rename(columns={0: "fof_secid", 1: "fof_sectype"})
    )
    if MASTER_PORTFOLIO_ID in df_top_level_holdings.columns:
        df_top_level_holdings["masterPortfolioId"] = df_top_level_holdings[MASTER_PORTFOLIO_ID].astype("Int64")

    query_str = """with cte as (
        SELECT distinct --performance_id,
        case
            when left(investment_id,2) ='E0' and fund_id is null
                then performance_id
            when is_primary=False
                then performance_id
            else investment_id
        end as sec_id
        ,investment_id
        ,morningstar_investment_name
        ,case
            when morningstar_instrument_type_code is not null
                then morningstar_instrument_type_code
            when left(investment_id,2) in ('FE','FC','FO','FH','FM','FV','SC')
                then left(investment_id,2)
            when left(investment_id,2) = 'F0'
                then 'FO'
            when left(investment_id,2) ='SA'
                then 'FS'
            when left(investment_id,2) ='E0' and fund_id is null
                then 'E'
            when left(investment_id,2) ='E0' and fund_id is not null
                then 'FC'
            else null
        end as morningstar_instrument_type_code
        ,is_primary
        FROM platform__reference__prd.investment_lookup a
        where
        (
            performance_id is not null
            or
            morningstar_instrument_type_code in ('BC','12','BZ','BY','B','BT','NE','BH','IP','NE','NB','NC','BG','BD','ND')
        )
        and
            left(investment_id,2) not in ('XI','VA')
        and
        (
            a.performance_id in (select fof_secid from temp_top_level_holdings_expanded)
            or
            a.investment_id in (select fof_secid from temp_top_level_holdings_expanded)
        )
    )
    select
        a.weight as weighting,
        a.Name as securityName,
        b.investment_id as secId,
        coalesce(trim(b.morningstar_instrument_type_code), 'Q') as detailHoldingTypeId
    from
        temp_top_level_holdings_expanded a
        left join cte b
            on a.fof_secid=b.sec_id
    """

    return datalake.query(
        query_str=query_str,
        temp_tables=[
            datalake.TempTable(name="temp_top_level_holdings_expanded", df=df_top_level_holdings),
        ],
    )


def _get_lookthrough_holdings_from_fof_api(portfolio_id: str, df_top_level_holdings: DataFrame) -> DataFrame:
    from ._backend_apis._signed_url_backend import SignedUrlBackend

    _signed_url_request = SignedUrlBackend()

    # 'XI' represents an index, which is unsupported by FoF API
    if "XI" in df_top_level_holdings["SecurityType"].values:
        raise BadRequestException(BAD_REQUEST_ERROR_INVALID_LOOKTHROUGH_HOLDING_TYPE)

    _logger.info(
        "Fetching FOF comptabile secid and investment types from investments_lookup table in lakehouse along with required fields from top level holdings to pass to FOF API."
    )
    df_fof_sec_id_investment_types = _get_fof_secid_and_investment_types(df_top_level_holdings)

    _logger.info("Generating a unique id for each row in the top level holdings.")
    # Add a UUID as value for clientSecurityGuid for each holding

    df_fof_sec_id_investment_types = df_fof_sec_id_investment_types.rename(
        columns={
            "secid": "secId",
            "weighting": "weighting",
            "detailholdingtypeid": "detailHoldingTypeId",
            "securityname": "securityName",
        }
    )

    df_fof_sec_id_investment_types["clientSecurityGuid"] = [
        str(uuid.uuid4()) for _ in range(len(df_fof_sec_id_investment_types.index))
    ]

    _logger.info("Extracting fof_secid, investment_type, weight and masterportfolio id into json to pass to FOF API")
    holdings_json = df_fof_sec_id_investment_types.to_json(orient="records")

    holdings_parsed = json.loads(holdings_json)  # This is sent to FOF API in the holdings field

    fof_post_url = f"{_config.fof_api_url()}portfolios/drilldown/{portfolio_id.split(';')[0]}/ALCustomerHoldings"
    payload = json.dumps(
        {
            "clientPortfolioGuid": str(uuid.uuid4()),
            "portfolioDate": str(datetime.now().strftime("%Y%m%d")),
            "currencyId": "USD",
            "holdings": holdings_parsed,  # The SQL query returns the columns in exactly the same format as required by FOF API
        }
    )

    _logger.info("Submitting request to FOF API to run lookthrough holdings.")
    _logger.debug(f"Sending request to {fof_post_url}")

    # MD proxy api will poll FOF API and return the signed url to download the lookthrough holdings
    signed_url = _fof_api_request.do_post_request(fof_post_url, payload)["jobId"]  # Returns reponse.content as string

    _logger.info("Submitting request to download lookthrough holdsings from signed url.")
    _logger.debug(f"Sending request to {signed_url}")
    look_through_holdings_reponse = _signed_url_request.do_get_request(
        signed_url
    )  # The reponse includes lookthrough holdings as json

    _logger.info("Got response from Signed url request")
    look_through_holdings_json = json.loads(look_through_holdings_reponse)
    holdings = look_through_holdings_json["holdings"]
    _logger.debug(f"Got {len(holdings)} lookthrough holdings.")

    df_look_through_holdings = pd.DataFrame(holdings)
    return df_look_through_holdings


def _has_ams_license_for_isin() -> bool:
    ams_get_url = f"{_config.al_proxy_api_url()}/v2/ams/api/user/entitlements"
    response = _ams_api_request.do_get_request(ams_get_url)

    data_groups = response["DataGroups"]
    for data_group in data_groups:
        if (
            (data_group["IdType"] == "UNIVERSEID" and data_group["DataGroupId"] == "ALL") or data_group["DataGroupId"] == "CUSIP"
        ) and data_group["Permission"] == 1:
            return True
    return False


# temporary function until this data is added to Lake House
def _morningstar_investment_type_id_name_mapping() -> pd.DataFrame:
    return pd.DataFrame(
        [
            {"morningstar_instrument_type_code": "0", "morningstar_instrument_type_name": "Muni Bond - Unspecified"},
            {"morningstar_instrument_type_code": "1", "morningstar_instrument_type_name": "Muni Bond - General Obligation"},
            {"morningstar_instrument_type_code": "12", "morningstar_instrument_type_name": "Muni Bond - Revenue"},
            {
                "morningstar_instrument_type_code": "13",
                "morningstar_instrument_type_name": "Muni Bond - Cash (Retired after Oct 2022)",
            },
            {"morningstar_instrument_type_code": "14", "morningstar_instrument_type_name": "Muni Bond - Double Barrelled"},
            {"morningstar_instrument_type_code": "AF", "morningstar_instrument_type_name": "Alternatives - Farm & Timber Land"},
            {"morningstar_instrument_type_code": "AI", "morningstar_instrument_type_name": "Alternatives - Infrastructure"},
            {
                "morningstar_instrument_type_code": "AM",
                "morningstar_instrument_type_name": "Alternatives - Master Investment Trust",
            },
            {"morningstar_instrument_type_code": "AP", "morningstar_instrument_type_name": "Alternatives - Private Equity"},
            {"morningstar_instrument_type_code": "AR", "morningstar_instrument_type_name": "Alternatives - Real Estate"},
            {"morningstar_instrument_type_code": "B", "morningstar_instrument_type_name": "Bond - Corporate Bond"},
            {"morningstar_instrument_type_code": "BB", "morningstar_instrument_type_name": "Bond - Short-term Corporate Bills"},
            {"morningstar_instrument_type_code": "BC", "morningstar_instrument_type_name": "Bond - Convertible"},
            {"morningstar_instrument_type_code": "BD", "morningstar_instrument_type_name": "Bond - Gov't Agency Debt"},
            {"morningstar_instrument_type_code": "BG", "morningstar_instrument_type_name": "Bond - Gov't Agency Pass-Thru"},
            {"morningstar_instrument_type_code": "BH", "morningstar_instrument_type_name": "Bond - Non-Agency Residential MBS"},
            {"morningstar_instrument_type_code": "BL", "morningstar_instrument_type_name": "Bond Index - Future"},
            {"morningstar_instrument_type_code": "BM", "morningstar_instrument_type_name": "Bond - Non-U.S. Gov't Agency MBS"},
            {
                "morningstar_instrument_type_code": "BO",
                "morningstar_instrument_type_name": "Bond Index - Option (Call) (Consolidation to FD after Oct 2022)",
            },
            {
                "morningstar_instrument_type_code": "BP",
                "morningstar_instrument_type_name": "Bond Index - Option (Put) (Consolidation to FD after Oct 2022)",
            },
            {"morningstar_instrument_type_code": "BQ", "morningstar_instrument_type_name": "Bond - Undefined"},
            {"morningstar_instrument_type_code": "BR", "morningstar_instrument_type_name": "Bond - Bank Loans"},
            {"morningstar_instrument_type_code": "BT", "morningstar_instrument_type_name": "Bond - Gov't/Treasury"},
            {"morningstar_instrument_type_code": "BU", "morningstar_instrument_type_name": "Bond - Units"},
            {"morningstar_instrument_type_code": "BW", "morningstar_instrument_type_name": "Bond - Warrants/Rights (Call)"},
            {"morningstar_instrument_type_code": "BX", "morningstar_instrument_type_name": "Bond - Warrants/Rights (Put)"},
            {"morningstar_instrument_type_code": "BY", "morningstar_instrument_type_name": "Bond - Asset Backed"},
            {"morningstar_instrument_type_code": "BZ", "morningstar_instrument_type_name": "Bond - Supranational"},
            {"morningstar_instrument_type_code": "C", "morningstar_instrument_type_name": "Cash"},
            {"morningstar_instrument_type_code": "CA", "morningstar_instrument_type_name": "Cash - Collateral"},
            {
                "morningstar_instrument_type_code": "CC",
                "morningstar_instrument_type_name": "Cash - Option (Call) (Consolidation to CY after Oct 2022)",
            },
            {"morningstar_instrument_type_code": "CD", "morningstar_instrument_type_name": "Cash - CD/Time Deposit"},
            {"morningstar_instrument_type_code": "CH", "morningstar_instrument_type_name": "Cash - Currency"},
            {"morningstar_instrument_type_code": "CL", "morningstar_instrument_type_name": "Currency - Future"},
            {"morningstar_instrument_type_code": "CN", "morningstar_instrument_type_name": "Bond - Contingent Convertible"},
            {
                "morningstar_instrument_type_code": "CO",
                "morningstar_instrument_type_name": "Cash - Option (Put) (Consolidation to CZ after Oct 2022)",
            },
            {"morningstar_instrument_type_code": "CP", "morningstar_instrument_type_name": "Cash - Commercial Paper"},
            {"morningstar_instrument_type_code": "CQ", "morningstar_instrument_type_name": "Cash - Future Offset"},
            {"morningstar_instrument_type_code": "CR", "morningstar_instrument_type_name": "Cash - Repurchase Agreement"},
            {"morningstar_instrument_type_code": "CS", "morningstar_instrument_type_name": "Currency - Swap"},
            {"morningstar_instrument_type_code": "CT", "morningstar_instrument_type_name": "Bond - Capital Contingent Debt"},
            {"morningstar_instrument_type_code": "CU", "morningstar_instrument_type_name": "Currency - Forward"},
            {
                "morningstar_instrument_type_code": "CV",
                "morningstar_instrument_type_name": "Currency - Warrants\\Rights (Call) (Consolidation to CY after Oct 2022)",
            },
            {
                "morningstar_instrument_type_code": "CX",
                "morningstar_instrument_type_name": "Currency - Warrants\\Rights (Put) (Consolidation to CZ after Oct 2022)",
            },
            {"morningstar_instrument_type_code": "CY", "morningstar_instrument_type_name": "Currency Option (Call)"},
            {"morningstar_instrument_type_code": "CZ", "morningstar_instrument_type_name": "Currency Option (Put)"},
            {"morningstar_instrument_type_code": "DA", "morningstar_instrument_type_name": "Bond - Future"},
            {"morningstar_instrument_type_code": "DB", "morningstar_instrument_type_name": "Bond - Option (Call)"},
            {"morningstar_instrument_type_code": "DC", "morningstar_instrument_type_name": "Commodity - Option (Call)"},
            {"morningstar_instrument_type_code": "DD", "morningstar_instrument_type_name": "Commodity"},
            {"morningstar_instrument_type_code": "DE", "morningstar_instrument_type_name": "Bond - Option (Put)"},
            {"morningstar_instrument_type_code": "DG", "morningstar_instrument_type_name": "Equity - Future"},
            {"morningstar_instrument_type_code": "DH", "morningstar_instrument_type_name": "Equity - Option (Call)"},
            {"morningstar_instrument_type_code": "DI", "morningstar_instrument_type_name": "Equity - Option (Put)"},
            {"morningstar_instrument_type_code": "DJ", "morningstar_instrument_type_name": "Other - Future"},
            {"morningstar_instrument_type_code": "DM", "morningstar_instrument_type_name": "Commodity - Future"},
            {
                "morningstar_instrument_type_code": "DO",
                "morningstar_instrument_type_name": "Bond - Collateralized Debt Obligations (CDO/CBO)",
            },
            {"morningstar_instrument_type_code": "DP", "morningstar_instrument_type_name": "Commodity - Option (Put)"},
            {
                "morningstar_instrument_type_code": "DS",
                "morningstar_instrument_type_name": "Bond - Sub-sovereign Government Debt",
            },
            {"morningstar_instrument_type_code": "E", "morningstar_instrument_type_name": "Equity"},
            {"morningstar_instrument_type_code": "EC", "morningstar_instrument_type_name": "Equity Index - Option (Call)"},
            {"morningstar_instrument_type_code": "EL", "morningstar_instrument_type_name": "Equity Index - Future"},
            {"morningstar_instrument_type_code": "EP", "morningstar_instrument_type_name": "Equity Index - Option (Put)"},
            {"morningstar_instrument_type_code": "EQ", "morningstar_instrument_type_name": "Equity - Undefined"},
            {"morningstar_instrument_type_code": "ER", "morningstar_instrument_type_name": "Equity - REIT"},
            {"morningstar_instrument_type_code": "EU", "morningstar_instrument_type_name": "Equity - Units"},
            {"morningstar_instrument_type_code": "EV", "morningstar_instrument_type_name": "Equity - Warrants/Rights (Put)"},
            {"morningstar_instrument_type_code": "EW", "morningstar_instrument_type_name": "Equity - Warrants/Rights (Call)"},
            {"morningstar_instrument_type_code": "EX", "morningstar_instrument_type_name": "Mutual Fund - Unspecified"},
            {"morningstar_instrument_type_code": "FC", "morningstar_instrument_type_name": "Mutual Fund - Closed End"},
            {"morningstar_instrument_type_code": "FD", "morningstar_instrument_type_name": "Other Fixed Income Derivative"},
            {"morningstar_instrument_type_code": "FE", "morningstar_instrument_type_name": "Mutual Fund - ETF"},
            {"morningstar_instrument_type_code": "FH", "morningstar_instrument_type_name": "Mutual Fund - Hedge Fund"},
            {"morningstar_instrument_type_code": "FM", "morningstar_instrument_type_name": "Mutual Fund - Money Market"},
            {"morningstar_instrument_type_code": "FO", "morningstar_instrument_type_name": "Mutual Fund - Open End"},
            {"morningstar_instrument_type_code": "FS", "morningstar_instrument_type_name": "Mutual Fund - Separate Account"},
            {"morningstar_instrument_type_code": "FV", "morningstar_instrument_type_name": "Mutual Fund - Variable Annuity"},
            {"morningstar_instrument_type_code": "FZ", "morningstar_instrument_type_name": "Mutual Fund - Client Portfolio"},
            {"morningstar_instrument_type_code": "GA", "morningstar_instrument_type_name": "Bond - Non-US Gov't Agency CMO"},
            {"morningstar_instrument_type_code": "GC", "morningstar_instrument_type_name": "Bond - Global Non-Agency CMO"},
            {"morningstar_instrument_type_code": "GM", "morningstar_instrument_type_name": "Bond - Global Non-Agency MBS"},
            {"morningstar_instrument_type_code": "GS", "morningstar_instrument_type_name": "Bond - Short-term Government Bills"},
            {"morningstar_instrument_type_code": "IP", "morningstar_instrument_type_name": "Bond - Corp Inflation Protected"},
            {"morningstar_instrument_type_code": "IS", "morningstar_instrument_type_name": "Inflation Swap"},
            {"morningstar_instrument_type_code": "IT", "morningstar_instrument_type_name": "Income Trust"},
            {
                "morningstar_instrument_type_code": "LO",
                "morningstar_instrument_type_name": "Bond - Collateralized Loan Obligations (CLO)",
            },
            {"morningstar_instrument_type_code": "NB", "morningstar_instrument_type_name": "Bond - Commercial MBS"},
            {"morningstar_instrument_type_code": "NC", "morningstar_instrument_type_name": "Bond - Gov't Agency CMO"},
            {"morningstar_instrument_type_code": "ND", "morningstar_instrument_type_name": "Bond - Covered Bond"},
            {"morningstar_instrument_type_code": "NE", "morningstar_instrument_type_name": "Bond - Gov't Agency ARM"},
            {"morningstar_instrument_type_code": "NR", "morningstar_instrument_type_name": "Bond - U.S. Agency Credit Risk CMO"},
            {"morningstar_instrument_type_code": "OO", "morningstar_instrument_type_name": "Cash - Option Offset"},
            {"morningstar_instrument_type_code": "OS", "morningstar_instrument_type_name": "Cash - Swap Offset"},
            {"morningstar_instrument_type_code": "OT", "morningstar_instrument_type_name": "Cash - Forward Offset"},
            {"morningstar_instrument_type_code": "P", "morningstar_instrument_type_name": "Preferred Stock"},
            {"morningstar_instrument_type_code": "PA", "morningstar_instrument_type_name": "Participating Preferred"},
            {"morningstar_instrument_type_code": "PC", "morningstar_instrument_type_name": "Convertible Preferred"},
            {"morningstar_instrument_type_code": "PP", "morningstar_instrument_type_name": "Property"},
            {"morningstar_instrument_type_code": "PS", "morningstar_instrument_type_name": "Interest Rate Swaption - Payer"},
            {"morningstar_instrument_type_code": "Q", "morningstar_instrument_type_name": "Unidentified Holding"},
            {"morningstar_instrument_type_code": "QQ", "morningstar_instrument_type_name": "Other Assets And Liabilities"},
            {"morningstar_instrument_type_code": "RS", "morningstar_instrument_type_name": "Interest Rate Swaption - Receiver"},
            {"morningstar_instrument_type_code": "SA", "morningstar_instrument_type_name": "Asset Swap"},
            {"morningstar_instrument_type_code": "SC", "morningstar_instrument_type_name": "Mutual Fund - CIT"},
            {"morningstar_instrument_type_code": "SD", "morningstar_instrument_type_name": "Debt Swap"},
            {"morningstar_instrument_type_code": "SE", "morningstar_instrument_type_name": "Equity Swap"},
            {"morningstar_instrument_type_code": "SF", "morningstar_instrument_type_name": "Contract For Difference"},
            {"morningstar_instrument_type_code": "SI", "morningstar_instrument_type_name": "Interest Rate Swap"},
            {"morningstar_instrument_type_code": "SJ", "morningstar_instrument_type_name": "Interest Rate Future"},
            {"morningstar_instrument_type_code": "SK", "morningstar_instrument_type_name": "Interest Rate Forward"},
            {"morningstar_instrument_type_code": "SN", "morningstar_instrument_type_name": "Volatility/Variance Swap"},
            {"morningstar_instrument_type_code": "SQ", "morningstar_instrument_type_name": "Equity Index Swap"},
            {"morningstar_instrument_type_code": "SR", "morningstar_instrument_type_name": "Credit Default Swap"},
            {"morningstar_instrument_type_code": "ST", "morningstar_instrument_type_name": "Total Return Swap"},
            {"morningstar_instrument_type_code": "SU", "morningstar_instrument_type_name": "Structured Product"},
            {"morningstar_instrument_type_code": "SV", "morningstar_instrument_type_name": "Cash - Stable Value Fund"},
            {"morningstar_instrument_type_code": "SW", "morningstar_instrument_type_name": "Credit Default Index Swap (CDX)"},
            {"morningstar_instrument_type_code": "SX", "morningstar_instrument_type_name": "Swaption - Payer"},
            {"morningstar_instrument_type_code": "SY", "morningstar_instrument_type_name": "Swaption - Receiver"},
            {"morningstar_instrument_type_code": "TF", "morningstar_instrument_type_name": "Bond - Treasury Future"},
            {"morningstar_instrument_type_code": "TG", "morningstar_instrument_type_name": "Bond - U.S. Agency TBA"},
            {"morningstar_instrument_type_code": "TP", "morningstar_instrument_type_name": "Bond - Gov't Inflation Protected"},
            {"morningstar_instrument_type_code": "WR", "morningstar_instrument_type_name": "Undefined - Warrants/Rights"},
            {"morningstar_instrument_type_code": "XP", "morningstar_instrument_type_name": "Credit Default Swaption - Payer"},
            {"morningstar_instrument_type_code": "XR", "morningstar_instrument_type_name": "Credit Default Swaption - Receiver"},
            {"morningstar_instrument_type_code": "YF", "morningstar_instrument_type_name": "Cryptocurrency - Future"},
            {"morningstar_instrument_type_code": "YO", "morningstar_instrument_type_name": "Cryptocurrency"},
        ]
    )


@_decorator.typechecked
def get_lookthrough_holdings(portfolio_id: str) -> DataFrame:
    """
    :bdg-ref-danger:`Upcoming Feature <../upcoming_feature>`

    Returns look-through holdings for a user-created portfolio.

    Args:
        portfolio_id (:obj:`str`): Portfolio ID. Use the `get_portfolios <../portfolio/portfolio_list.html#morningstar_data.direct.user_items.get_portfolios>`_ function to discover saved portfolios.

    :Returns:
        DataFrame: A DataFrame object with holdings data. DataFrame columns include

        * investment_id
        * isin
        * cusip
        * weight
        * market_value
        * security_name
        * currency
        * morningstar_instrument_type_code
        * entity_id
        * morningstar_entity_name
        * primary_performance_id
        * morningstar_instrument_type_name

    :Examples:

    Retrieve the look-through holdings for a portfolio.
    ::

        import morningstar_data as md

        df = md.direct.get_lookthrough_holdings(portfolio_id='7b9cb5db-e3da-414e-8f75-b52b02222b5a') # Replace with a valid Portfolio ID
        df

    :Output:

        =============  === ======================  ================================
        investment_id  ... primary_performance_id  morningstar_instrument_type_name
        =============  === ======================  ================================
        E0USA00462         0P000000YM              Equity
        F00000Q5HQ	       0P0000Z4DP              Equity
        F00000UMZ1         0P00014J84              Equity
        B10002POBD         None                    Bond - Asset Backed
        =============  === ======================  ================================

    :Errors:
        InternalServerError: Raised when the server encounters an unhandled error.

        NetworkExceptionError: Raised when the request fails to reach the server due to a network error.

        TimeoutError: Raised when the fund of fund calculation takes too long.

        ResourceNotFoundError: Raised when portfolio_id does not exist in Direct.

        BadRequestException: Raised when portfolio_id is an invalid UUID.

        UnavailableExternally: Raised when the function is not available for external Python package callers.
    """

    _logger.info("morningstar_data.direct.get_lookthrough_holdings")

    # Get top level holdings in the portfolio
    _logger.info("Fetching top level holdings")

    # Split portfolio type from porfolio id
    portfolio_id = portfolio_id.split(";")[0]
    try:
        uuid.UUID(portfolio_id)
    except ValueError:
        raise BadRequestException(BAD_REQUEST_ERROR_INVALID_PORTFOLIO_ID) from None

    df_top_level_holdings = portfolio.get_holdings(portfolio_ids=[portfolio_id])

    # Convert the weight to float, bcoz we think the FOF requires weight as float
    df_top_level_holdings["Weight"] = df_top_level_holdings["Weight"].astype(float)

    # Split secid and security type. They are bound together by ;
    df_top_level_holdings[["SecId", "SecurityType"]] = df_top_level_holdings["HoldingId"].str.split(";", expand=True)

    # We only are going to query FO and FE fund types. (open ended funds and ETFs)
    # Maybe we need all fund types ???
    fund_secids = df_top_level_holdings.loc[df_top_level_holdings["SecurityType"].isin(["FO", "FE"])]["SecId"].tolist()

    investments = fund_secids

    # Get investment data for each top level fund holding in the portfolio
    # We are making this call to get  masterportfolio_id.
    # If masterportfolio_id is added to lakehouse, this step can be removed

    # only if a portfolio has funds of type FO or FE find the MasterPortfolio Id
    if investments:
        _logger.info("Fetching investment data for all top level holdings")
        _logger.debug(f"{investments}")

        data_points = [
            {"datapointId": "DC09A", "datapointName": MASTER_PORTFOLIO_ID},
            {"datapointId": "OS00I", "datapointName": "SecId"},
        ]
        df_mpid_mapping = inv.get_investment_data(investments, data_points)
        df_top_level_holdings = df_top_level_holdings.merge(
            df_mpid_mapping[["SecId", MASTER_PORTFOLIO_ID]], on="SecId", how="left"
        )

    _logger.info("Fetching lookthrough holdings from FOF API.")
    try:
        df_lookthrough = _get_lookthrough_holdings_from_fof_api(portfolio_id, df_top_level_holdings)
    except Exception as e:
        _logger.error(e)
        if type(e).__name__ == "InvalidQueryException":
            raise BadRequestException(f"Sorry, something went wrong on our end: {str(e)} Please contact AL support") from None
        else:
            raise e from None

    df_lookthrough = df_lookthrough.rename(
        columns={
            "securityName": "security_name",
            "holdingDetailId": "investment_id",
            "detailHoldingTypeId": "morningstar_instrument_type_code",
            "currencyId": "currency",
            "fractionalWeight": "weight",
        }
    )

    _logger.info("Fetching enriched info with company information.")
    df_enriched_data = _get_ids_from_lake_house(df_lookthrough)

    # add morningstar_instrument_type_name as new column
    df_enriched_data = df_enriched_data.merge(
        _morningstar_investment_type_id_name_mapping(), on="morningstar_instrument_type_code", how="left"
    )

    # If the user has AMS license, we return ISIN.
    if _has_ams_license_for_isin() == False:
        df_enriched_data = df_enriched_data.drop(["isin", "cusip"])

    return df_enriched_data



from string import Template
from typing import Optional

from . import _error_messages


class AccessDeniedError(Exception):
    def __init__(self, message: Optional[str] = None) -> None:
        if not message:
            super().__init__(_error_messages.ACCESS_DENIED_ERROR)
        else:
            super().__init__(message)


class ApiRequestException(Exception):
    def __init__(self, message: str) -> None:
        super().__init__(message)


class ApiResponseException(Exception):
    def __init__(self, message: str, status: int = 404) -> None:
        self.status_code = status
        super().__init__(message)


class BadRequestException(Exception):
    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(message)


class CredentialsException(Exception):
    def __init__(self, message: str) -> None:
        super().__init__(message)


class ConnectionException(Exception):
    def __init__(self, message: str) -> None:
        super().__init__(message)


class ClientError(Exception):
    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(message)


class ForbiddenError(Exception):
    def __init__(self, message: Optional[str] = None) -> None:
        if not message:
            super().__init__(_error_messages.FORBIDDEN_ERROR)
        else:
            super().__init__(message)


class InvalidQueryException(Exception):
    def __init__(self, message: str) -> None:
        super().__init__(message)


class InternalServerError(Exception):
    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(message)


class NetworkExceptionError(Exception):
    def __init__(self, message: Optional[str] = None) -> None:
        if not message:
            super().__init__(_error_messages.NETWORK_ERROR)
        else:
            super().__init__(message)


class QueryLimitException(Exception):
    def __init__(self, query_limit: Optional[str] = None, status: int = 403) -> None:
        self.status_code = status
        if query_limit is not None:
            tpl = Template(_error_messages.QUERY_LIMIT_ERROR_SHOW_LIMIT)
            super().__init__(tpl.substitute(query=query_limit))
        else:
            super().__init__(_error_messages.QUERY_LIMIT_ERROR)


class ResourceNotFoundError(Exception):
    def __init__(self, message: Optional[str] = None) -> None:
        if not message:
            super().__init__(_error_messages.RESOURCE_NOT_FOUND_ERROR)
        else:
            super().__init__(message)


class TimeoutError(Exception):
    def __init__(self, message: Optional[str] = None) -> None:
        if not message:
            super().__init__(_error_messages.TIMEOUT_ERROR)
        else:
            super().__init__(message)


class ValueErrorException(Exception):
    def __init__(self, message: str) -> None:
        super().__init__(message)


class EAMSTokenDoesNotExistError(Exception):
    def __init__(self, message: str) -> None:
        super().__init__(message)


class MalformedJWTError(Exception):
    def __init__(self, message: Optional[str] = None) -> None:
        if not message:
            super().__init__(_error_messages.MALFORMED_JWT_ERROR)
        else:
            super().__init__(message)


class UnauthorizedDeliveryException(Exception):
    def __init__(self, message: str = _error_messages.DELIVERY_ACCESS_ERROR) -> None:
        super().__init__(message)


class FailedDeliveryError(Exception):
    def __init__(self, message: Optional[str] = None) -> None:
        if not message:
            super().__init__(_error_messages.FAILED_DELIVERY_ERROR)
        else:
            super().__init__(message)
